from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, DefaultDict

from better_translation._babel.lazy_proxy import LazyProxy
from better_translation.types import (
    Locale,
    MessageID,
    TranslatedText,
)

if TYPE_CHECKING:
    from collections.abc import Callable

    from better_translation.translator import ITranslator


@dataclass(slots=True)
class BaseTextProvider(ABC):
    @abstractmethod
    def _gettext(
        self,
        message_id: MessageID,
        /,
        *,
        default: TranslatedText | None = None,
        default_plural: TranslatedText | None = None,
        context: Any | None = None,
        locale: Locale | None = None,
        n: int | None = None,
        **extra: Any,
    ) -> str:
        ...

    def gettext(
        self,
        message_id: MessageID,
        /,
        *,
        default: TranslatedText | None = None,
        context: Any | None = None,
        locale: Locale | None = None,
        **extra: Any,
    ) -> str:
        return self._gettext(
            message_id,
            default=default,
            context=context,
            locale=locale,
            **extra,
        )

    def ngettext(
        self,
        message_id: MessageID,
        /,
        *,
        n: int,
        default: TranslatedText | None = None,
        default_plural: TranslatedText | None = None,
        context: Any | None = None,
        locale: Locale | None = None,
        **extra: Any,
    ) -> str:
        return self._gettext(
            message_id,
            default=default,
            default_plural=default_plural,
            context=context,
            locale=locale,
            n=n,
            **extra,
        )

    def lazy_gettext(
        self,
        message_id: MessageID,
        /,
        *,
        default: TranslatedText | None = None,
        context: Any | None = None,
        locale: Locale | None = None,
        **extra: Any,
    ) -> LazyProxy:
        return LazyProxy(
            self.gettext,
            message_id,
            default=default,
            context=context,
            locale=locale,
            **extra,
        )

    def lazy_ngettext(
        self,
        message_id: MessageID,
        /,
        *,
        n: int,
        default: TranslatedText | None = None,
        default_plural: TranslatedText | None = None,
        context: Any | None = None,
        locale: Locale | None = None,
        **extra: Any,
    ) -> LazyProxy:
        return LazyProxy(
            self.ngettext,
            message_id,
            default=default,
            default_plural=default_plural,
            context=context,
            locale=locale,
            n=n,
            **extra,
        )


@dataclass(slots=True)
class TextProvider(BaseTextProvider):
    translators: DefaultDict[Locale, ITranslator]
    locale_getter: Callable[[], Locale] = lambda: Locale("en")  # noqa: E731
    """Function to get current locale in case it is not provided."""

    def _gettext(
        self,
        message_id: MessageID,
        /,
        *,
        default: TranslatedText | None = None,
        default_plural: TranslatedText | None = None,
        context: Any | None = None,  # noqa: ARG002
        locale: Locale | None = None,
        n: int | None = None,
        **_: Any,
    ) -> TranslatedText:
        if locale is None:
            locale = self.locale_getter()

        translator = self.translators[locale]
        return translator.translate(
            message_id=message_id,
            locale=locale,
            default=default,
            default_plural=default_plural,
            n=n,
        )



from __future__ import annotations

from typing import cast

try:
    from django.db import models
except ImportError as e:
    msg = "Django is not installed. To use Django integration, install Django."
    raise ValueError(msg) from e


class DefaultLocalesChoices(models.TextChoices):
    RU = "ru", "Russian"
    EN = "en", "English"


class BaseMessage(models.Model):
    id = models.CharField(max_length=255, primary_key=True)
    default = models.TextField(verbose_name="Default singular form")
    default_plural = models.TextField(verbose_name="Default plural form")
    context = models.CharField(
        max_length=1024,
        default="",
        blank=True,
    )
    has_plural = models.BooleanField(verbose_name="Has plural", default=False)
    is_used = models.BooleanField(verbose_name="Is used", default=True)

    translations: models.QuerySet[BaseTranslation]
    """A QuerySet of translations for this message."""

    def __str__(self) -> str:
        return f"Message: {self.id}"

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, BaseMessage):
            return False

        return self.id == other.id

    def __hash__(self) -> int:
        return hash(self.id)

    class _Meta:
        abstract = True

    Meta = cast("type[models.Model.Meta]", _Meta)


class BaseTranslation(models.Model):
    locale = models.CharField(
        max_length=5,
        verbose_name="Locale",
        choices=DefaultLocalesChoices.choices,
    )
    singular = models.TextField(verbose_name="Singular form of the text")
    plural = models.TextField(
        verbose_name="Plural form of the text",
        default="",
        blank=True,
    )

    message = models.ForeignKey(
        BaseMessage,
        on_delete=models.CASCADE,
        related_name="translations",
    )

    def __str__(self) -> str:
        return f"MessageTranslation('{self.locale}', '{self.singular[:10]}...')"

    class _Meta:
        abstract = True

    Meta = cast("type[models.Model.Meta]", _Meta)



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from django import forms
from django.contrib import admin
from django.db import models

from better_translation.integrations.django.models import (
    BaseMessage,
    BaseTranslation,
)

if TYPE_CHECKING:
    from collections.abc import Callable, Sequence

    from django.http import HttpRequest

admin.ModelAdmin.__class_getitem__ = classmethod(  # type: ignore[attr-defined]
    lambda cls, _: cls,
)
admin.TabularInline.__class_getitem__ = classmethod(  # type: ignore[attr-defined]
    lambda cls, _: cls,
)


class BaseMessageAdmin(admin.ModelAdmin[BaseMessage]):
    list_display = ("id", "default", "context", "is_used")
    readonly_fields = ("id", "default", "default_plural", "context", "is_used")
    search_fields = (
        "id",
        "default",
        "default_plural",
        "context",
        "translations__singular",
        "translations__plural",
    )

    def get_fields(
        self,
        request: HttpRequest,  # noqa: ARG002
        obj: BaseMessage | None = None,
    ) -> Sequence[Callable[..., Any] | str]:
        if obj is None:
            return ("id", "default", "default_plural", "context", "is_used")

        fields = ["id", "default"]
        if obj.default_plural != obj.default:
            fields.append("default_plural")

        if obj.context:
            fields.append("context")

        fields.append("is_used")

        return fields


class BaseTranslationInline(admin.TabularInline[BaseTranslation]):
    model = BaseTranslation
    extra = 0
    formfield_overrides = {
        models.TextField: {
            "widget": forms.Textarea(attrs={"rows": 6, "cols": 50}),
        },
    }

    # pyright: reportIncompatibleMethodOverride=false
    def get_fields(
        self,
        request: HttpRequest,  # noqa: ARG002
        obj: BaseMessage | None = None,
    ) -> Sequence[Callable[..., Any] | str]:
        if obj is None or obj.has_plural:
            return ("locale", "singular", "plural")

        return ("locale", "singular")



from __future__ import annotations

import logging
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import TYPE_CHECKING

from better_translation.storage import StorageMessage

if TYPE_CHECKING:
    from better_translation.storage import IStorage
    from better_translation.types import (
        Locale,
        MessageID,
        TranslatedText,
    )


logger = logging.getLogger(__name__)


class ITranslator(ABC):
    @abstractmethod
    def translate(  # noqa: PLR0913
        self,
        message_id: MessageID,
        locale: Locale,
        default: TranslatedText | None = None,
        default_plural: TranslatedText | None = None,
        n: int | None = None,
    ) -> TranslatedText:
        ...


@dataclass(slots=True)
class DefaultTranslator(ITranslator):
    storage: IStorage

    def translate(  # noqa: PLR0913
        self,
        message_id: MessageID,
        locale: Locale,
        default: TranslatedText | None = None,
        default_plural: TranslatedText | None = None,
        n: int | None = None,
    ) -> TranslatedText:
        if n is None:
            return self._translate_singular(
                message_id=message_id,
                locale=locale,
                default=default,
            )

        return self._translate_plural(
            message_id=message_id,
            locale=locale,
            n=n,
            default=default,
            default_plural=default_plural,
        )

    def _translate_singular(
        self,
        message_id: MessageID,
        locale: Locale,
        default: TranslatedText | None = None,
    ) -> TranslatedText:
        message = self._get_message(
            message_id=message_id,
            default=default,
        )

        translation = message.translations.get(locale)
        if translation is not None:
            return translation.singular

        return message.default

    def _translate_plural(  # noqa: PLR0913
        self,
        message_id: MessageID,
        locale: Locale,
        n: int,
        default: TranslatedText | None = None,
        default_plural: str | None = None,
    ) -> TranslatedText:
        message = self._get_message(
            message_id=message_id,
            default=default,
            default_plural=default_plural,
        )

        translation = message.translations.get(locale)
        if translation is not None:
            singular, plural = translation.singular, translation.plural
        else:
            singular, plural = message.default, message.default_plural

        return self._get_singular_or_plural(n, singular, plural)

    def _get_message(
        self,
        message_id: MessageID,
        default: TranslatedText | None = None,
        default_plural: TranslatedText | None = None,
    ) -> StorageMessage:
        message = self.storage.get_message(message_id=message_id)
        if message is not None:
            return message

        default = default or message_id
        default_plural = default_plural or message_id

        logger.warning(
            "No message found for '%s'. Using the default values: '%s', '%s'",
            message_id,
            default,
            default_plural,
        )

        return StorageMessage(
            id=message_id,
            default=default,
            default_plural=default_plural,
            translations={},
            context="<no context>",
            has_plural=False,
        )

    def _get_singular_or_plural(
        self,
        n: int,
        singular: TranslatedText,
        plural: TranslatedText,
    ) -> TranslatedText:
        if n == 1:
            return singular

        return plural



from __future__ import annotations

import re
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from collections.abc import Iterable, Sequence
    from pathlib import Path

    from better_translation.types import MessageID, TranslatedText


FUNCTION_NAMES_TO_EXTRACT_FROM = (
    "gettext",
    "ngettext",
    "lazy_gettext",
    "lazy_ngettext",
    "_",
    "__",
)


def _get_regex_to_extract_function_calls(
    function_names: Sequence[str],
) -> re.Pattern[str]:
    """Return a regex that matches calls of the given function names."""
    _function_names = "|".join(function_names)
    return re.compile(rf"""({_function_names})(\(["|'].*["|'].*\))""")


@dataclass(slots=True)
class ExtractedMessage:
    id: MessageID = field(init=False)
    default: TranslatedText | None = field(init=False)
    default_plural: TranslatedText | None = field(init=False)
    context: Any | None = field(init=False)
    extra: dict[str, str | int | None] = field(init=False)

    def __init__(
        self,
        message_id: str,
        /,
        default: TranslatedText | None = None,
        default_plural: TranslatedText | None = None,
        context: Any | None = None,
        **extra: Any,
    ) -> None:
        self.id = message_id
        self.default = default
        self.default_plural = default_plural
        self.context = context
        self.extra = extra


def extract_from_code(
    code: str,
    function_names: Sequence[str] = FUNCTION_NAMES_TO_EXTRACT_FROM,
) -> Iterable[ExtractedMessage]:
    """Extract messages from Python code."""
    regex = _get_regex_to_extract_function_calls(function_names)
    for match in regex.finditer(code):
        call_params = match.group(2)
        yield eval(f"ExtractedMessage{call_params}")  # noqa: PGH001


def extract_from_file(
    file_path: Path,
    function_names: Sequence[str] = FUNCTION_NAMES_TO_EXTRACT_FROM,
) -> Iterable[ExtractedMessage]:
    """Extract messages from a Python file."""
    yield from extract_from_code(file_path.read_text(), function_names)


def extract_from_dir(
    directory_path: Path,
    function_names: Sequence[str] = FUNCTION_NAMES_TO_EXTRACT_FROM,
) -> Iterable[ExtractedMessage]:
    """Extract messages from Python files in a directory."""
    for path in directory_path.rglob("*.py"):
        yield from extract_from_file(path, function_names)



from __future__ import annotations

import logging
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import TYPE_CHECKING

from better_translation.extractor import extract_from_dir

if TYPE_CHECKING:
    from collections.abc import Iterable, Mapping
    from pathlib import Path

    from better_translation.types import (
        Locale,
        MessageID,
        TranslatedText,
    )

logger = logging.getLogger(__name__)


@dataclass(slots=True)
class StorageMessage:
    id: MessageID
    default: TranslatedText
    default_plural: TranslatedText
    context: str
    has_plural: bool
    translations: Mapping[Locale, StorageTranslation]

    @property
    def number_of_non_default_attributes(self) -> int:
        number_of_non_default_attributes = 1

        if self.default != self.id:
            number_of_non_default_attributes += 1

        if self.default_plural != self.default:
            number_of_non_default_attributes += 1

        if self.context:
            number_of_non_default_attributes += 1

        if self.has_plural:
            number_of_non_default_attributes += 1

        return number_of_non_default_attributes


@dataclass(slots=True)
class StorageTranslation:
    singular: TranslatedText
    plural: TranslatedText


class IStorage(ABC):
    @abstractmethod
    def get_message(self, message_id: MessageID) -> StorageMessage | None:
        """Get a message from the storage."""

    @abstractmethod
    async def load(self) -> None:
        """Load translations from the storage to the memory."""

    @abstractmethod
    async def update_messages(
        self,
        directory_path: Path,
    ) -> None:
        """Update untranslated messages from the directory in the storage."""


@dataclass(slots=True)
class BaseStorage(IStorage, ABC):
    storage: dict[MessageID, StorageMessage] = field(
        default_factory=dict,
        init=False,
    )
    is_loaded: bool = field(default=False, init=False)

    def get_message(self, message_id: MessageID) -> StorageMessage | None:
        if not self.is_loaded:
            msg = "Cannot get translation, storage is not loaded..."
            raise ValueError(msg)

        return self.storage.get(message_id)

    @staticmethod
    def _extract_messages(directory_path: Path) -> Iterable[StorageMessage]:
        for message in extract_from_dir(directory_path):
            yield StorageMessage(
                id=message.id,
                default=message.default or message.id,
                default_plural=message.default_plural or message.id,
                context=str(message.context) if message.context else "",
                has_plural="n" in message.extra,
                translations={},
            )



from __future__ import annotations

import logging
from dataclasses import dataclass
from typing import TYPE_CHECKING

from better_translation.integrations.django.models import BaseMessage
from better_translation.storage import (
    BaseStorage,
    StorageMessage,
    StorageTranslation,
)

if TYPE_CHECKING:
    from pathlib import Path


logger = logging.getLogger(__name__)


@dataclass(slots=True)
class DjangoStorage(BaseStorage):
    message_model: type[BaseMessage] = BaseMessage

    def __post_init__(self) -> None:
        if self.message_model is BaseMessage:
            msg = "`message_model` is required"
            raise ValueError(msg)

    async def load(self) -> None:
        """Load translations from the storage to the memory."""
        logger.info("Loading messages from the database...")

        messages = self.message_model.objects.all().prefetch_related(
            "translations",
        )
        async for message in messages:
            translations = message.translations.all()
            if not translations:
                logger.warning(
                    "Message '%s' has no translations",
                    message,
                )
            else:
                logger.debug(
                    "Message '%s' has '%s' translations",
                    message,
                    len(translations),
                )

            self.storage[message.id] = StorageMessage(
                id=message.id,
                default=message.default,
                default_plural=message.default_plural,
                context=message.context,
                has_plural=message.has_plural,
                translations={
                    translation.locale: StorageTranslation(
                        singular=translation.singular,
                        plural=translation.singular,
                    )
                    for translation in translations
                },
            )

        self.is_loaded = True

        logger.info("Messages loaded successfully")

    async def update_messages(
        self,
        directory_path: Path,
    ) -> None:
        if not self.is_loaded:
            logger.info("Cannot update messages, storage is not loaded...")
            await self.load()

        logger.info("Updating messages from the directory...")

        new_messages: set[BaseMessage] = set()
        unused_messages = set(self.storage)

        messages = self._extract_messages(directory_path)
        for message in messages:
            if message.id in self.storage:
                unused_messages.remove(message.id)

                saved_message = self.storage[message.id]
                if (
                    message.number_of_non_default_attributes
                    >= saved_message.number_of_non_default_attributes
                ):
                    logger.debug(
                        "Message '%s' has more non-default attributes, updating...",
                        message,
                    )
                    self.storage[message.id] = message

                continue

            logger.debug(
                "Message '%s' is new, adding to the database...",
                message,
            )
            new_messages.add(self._init_message_model(message))
            self.storage[message.id] = message

        await self.message_model.objects.abulk_create(new_messages)
        updated_messages_count: int = (
            # pyright: reportGeneralTypeIssues=false
            await self.message_model.objects.abulk_update(
                (
                    self._init_message_model(message)
                    for message in self.storage.values()
                ),
                fields=(
                    "default",
                    "default_plural",
                    "has_plural",
                    "context",
                ),
            )
        )
        await self.message_model.objects.filter(
            id__in=unused_messages,
        ).aupdate(is_used=False)

        logger.info(
            "Updated %s messages, created %s messages, found %s unused messages",
            updated_messages_count,
            len(new_messages),
            len(unused_messages),
        )

    def _init_message_model(
        self,
        storage_message: StorageMessage,
    ) -> BaseMessage:
        return self.message_model(
            id=storage_message.id,
            default=storage_message.default,
            default_plural=storage_message.default_plural,
            context=storage_message.context,
            has_plural=storage_message.has_plural,
        )
